=begin
 This file is part of Hopsworks
 Copyright (C) 2024, Hopsworks AB. All rights reserved

 Hopsworks is free software: you can redistribute it and/or modify it under the terms of
 the GNU Affero General Public License as published by the Free Software Foundation,
 either version 3 of the License, or (at your option) any later version.

 Hopsworks is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 PURPOSE.  See the GNU Affero General Public License for more details.

 You should have received a copy of the GNU Affero General Public License along with this program.
 If not, see <https://www.gnu.org/licenses/>.
=end

describe "On #{ENV['OS']}" do
  after(:all) {clean_all_test_projects(spec: "environment")}

  describe 'job' do
    context 'without authentication' do
      before :all do
        with_valid_project
        reset_session
      end
      it "should fail" do
        get_environment(@project[:id], "spark-feature-pipeline", expected_status: 401, error_code: 200003)
      end
    end
    context 'with authentication create, delete, get' do
      before :all do
        with_valid_project
      end
      it "should fail to install library in base environment" do
        install_library(@project[:id], "python-feature-pipeline", 'imageio', 'PIP', nil, lib_version: '2.9.0', expected_status: 403)
      end
      it "should fail to delete base environment" do
        get_env(@project[:id], "python-feature-pipeline", expected_status: 200)
        delete_env(@project[:id], "python-feature-pipeline", expected_status: 403)
      end
      it "should fail to create with name containing uppercase characters" do
        create_environment(@project[:id], "Abc", "python-feature-pipeline", "", expected_status: 400, error_code: 300023)
      end
      it "should fail to create with name ending in separator" do
        create_environment(@project[:id], "a-b-", "python-feature-pipeline", "", expected_status: 400, error_code: 300023)
      end
      it "should fail to create with name starting with separator" do
        create_environment(@project[:id], "-a-b", "python-feature-pipeline", "", expected_status: 400, error_code: 300023)
      end
      it "should fail to create with name containing double underscore separator" do
        create_environment(@project[:id], "ab__c", "python-feature-pipeline", "", expected_status: 400, error_code: 300023)
      end
      it "should fail to create with name containing double dash separator" do
        create_environment(@project[:id], "1--3", "python-feature-pipeline", "", expected_status: 400, error_code: 300023)
      end    
      it "should get default environments" do
        get_environments(@project[:id])
        expect(json_body[:items].count).to eq 9
        expect(json_body[:count]).to eq 9
      end
      it "should get all existing environments" do
        get_environments(@project[:id])
        json_body[:items].each {|env|

        expect(env[:isPreinstalledDockerImage]).to be true
        expect(env[:isOldDockerImage]).to be false

        get_environment(@project[:id], env[:name])
        }
      end
      it "should create new environment from all default environments" do
        get_environments(@project[:id])
        expect(json_body[:items].count).to eq 9
        expect(json_body[:count]).to eq 9

        json_body[:items].each {|env|
        create_environment(@project[:id], "#{env[:name]}-new", "#{env[:name]}", "#{env[:name]} is the name of the env")
        }

        get_environments(@project[:id])
        expect(json_body[:items].count).to eq 18
        expect(json_body[:count]).to eq 18

        wait_for_running_command(@project[:id])

        get_environments(@project[:id])
        expect(json_body[:items].count).to eq 18
        expect(json_body[:count]).to eq 18

        get_environment(@project[:id], "python-feature-pipeline-new")
        expect(json_body[:isPreinstalledDockerImage]).to be false
        expect(json_body[:isOldDockerImage]).to be false
      end
      it "should install library concurrently" do
        install_library(@project[:id], "python-feature-pipeline-new", 'imageio', 'PIP', nil, lib_version: '2.9.0')
        install_library(@project[:id], "spark-feature-pipeline-new", 'imageio', 'PIP', nil, lib_version: '2.10.1')
        install_library(@project[:id], "torch-training-pipeline-new", 'imageio', 'PIP', nil, lib_version: '2.10.2')

        install_library(@project[:id], "tensorflow-training-pipeline-new", 'imageio', 'PIP', nil, lib_version: '2.10.3')
        uninstall_library(@project[:id], "tensorflow-training-pipeline-new", 'polars')
        install_library(@project[:id], "tensorflow-training-pipeline-new", 'polars', 'PIP', nil, lib_version: '0.20.7')

        install_library(@project[:id], "misc-training-pipeline-new", 'imageio', 'PIP', nil, lib_version: '2.10.4')
        install_library(@project[:id], "minimal-inference-pipeline-new", 'imageio', 'PIP', nil, lib_version: '2.10.5')

        wait_for_running_command(@project[:id])

        list_libraries(@project[:id], "python-feature-pipeline-new")
        imageio_library = json_body[:items].detect { |library| library[:library] == "imageio" }
        expect(imageio_library).not_to be_nil
        expect(imageio_library[:version]).to eq ('2.9.0')

        list_libraries(@project[:id], "spark-feature-pipeline-new")
        imageio_library = json_body[:items].detect { |library| library[:library] == "imageio" }
        expect(imageio_library).not_to be_nil
        expect(imageio_library[:version]).to eq ('2.10.1')

        list_libraries(@project[:id], "torch-training-pipeline-new")
        imageio_library = json_body[:items].detect { |library| library[:library] == "imageio" }
        expect(imageio_library).not_to be_nil
        expect(imageio_library[:version]).to eq ('2.10.2')

        list_libraries(@project[:id], "tensorflow-training-pipeline-new")
        imageio_library = json_body[:items].detect { |library| library[:library] == "imageio" }
        polars_library = json_body[:items].detect { |library| library[:library] == "polars" }
        expect(imageio_library).not_to be_nil
        expect(imageio_library[:version]).to eq ('2.10.3')
        expect(polars_library).not_to be_nil
        expect(polars_library[:version]).to eq ('0.20.7')

        list_libraries(@project[:id], "misc-training-pipeline-new")
        imageio_library = json_body[:items].detect { |library| library[:library] == "imageio" }
        expect(imageio_library).not_to be_nil
        expect(imageio_library[:version]).to eq ('2.10.4')
                                
        list_libraries(@project[:id], "minimal-inference-pipeline-new")
        imageio_library = json_body[:items].detect { |library| library[:library] == "imageio" }
        expect(imageio_library).not_to be_nil
        expect(imageio_library[:version]).to eq ('2.10.5')
      end
      it "should filter on name" do
        get_environments(@project[:id], query: "?filter_by=name:minimal-inference-pipeline-new")
        expect(json_body[:items].count).to eq 1
        expect(json_body[:count]).to eq 1

        get_environments(@project[:id], query: "?filter_by=name:inference-pipeline")
        expect(json_body[:items].count).to eq 8
        expect(json_body[:count]).to eq 8
      end
      it "should filter on description" do
        get_environments(@project[:id], query: "?filter_by=description:tensorflow")
        expect(json_body[:items].count).to eq 2
        expect(json_body[:count]).to eq 2

        get_environments(@project[:id], query: "?filter_by=description:not_here")
        expect(json_body[:items]).to be_nil
        expect(json_body[:count]).to eq 0
      end
      it "should filter on base_image_name" do
        get_environments(@project[:id], query: "?filter_by=base_image_name:python-feature-pipeline")
        expect(json_body[:items].count).to eq 2
        expect(json_body[:count]).to eq 2
      end
      it "should filter on base_image_type" do
        get_environments(@project[:id], query: "?filter_by=base_image_type:feature_pipeline")
        expect(json_body[:items].count).to eq 4
        expect(json_body[:count]).to eq 4
      end
      describe 'with different environments' do
        context 'for jupyter' do
          before :all do
            with_valid_project
          end
          it "should check default environment" do
            get_settings(@project)
            expect(json_body[:environment][:name]).to eq "spark-feature-pipeline"
          end
          it "should be able to change environment" do
            get_settings(@project)
            settings = json_body
            settings[:environment][:name] = "python-feature-pipeline"
            start_jupyter(@project, settings: settings)
            get_settings(@project)
            expect(json_body[:environment][:name]).to eq "python-feature-pipeline"
            stop_jupyter(@project)
          end
          it "should fail when configuring to non-existent environment" do
            get_settings(@project)
            settings = json_body
            settings[:environment][:name] = "does_not_exist"
            start_jupyter(@project, settings: settings, expected_status: 404, error_code: 300000)
          end
          it "should fail when configuring to not supported environment" do
            get_settings(@project)
            settings = json_body
            settings[:environment][:name] = "tensorflow-inference-pipeline"
            start_jupyter(@project, settings: settings, expected_status: 400, error_code: 300021)
          end
          it "should fail when trying to remove environment in use" do
            create_environment(@project[:id], "jupyter_env", "python-feature-pipeline", "some description")
            wait_for_running_command(@project[:id])

            get_settings(@project)
            settings = json_body
            settings[:environment][:name] = "jupyter_env"
            start_jupyter(@project, settings: settings)
            stop_jupyter(@project)

            delete_env(@project[:id], "jupyter_env", expected_status: 400, error_code: 300022)

            settings[:environment][:name] = "spark-feature-pipeline"
            start_jupyter(@project, settings: settings)
            stop_jupyter(@project)

            delete_env(@project[:id], "jupyter_env")
            expect_status_details(204)
          end
        end
        context 'for job' do
          before :all do
            if ENV['OS'] == "ubuntu"
              skip "These tests do not run on ubuntu"
            end
            with_valid_project
          end
          it "check default python job image" do
            get_default_job_configuration(@project[:id], "python")
            expect(json_body[:environmentName]).to eq "python-feature-pipeline"
          end
          it "check default pyspark job image" do
            get_default_job_configuration(@project[:id], "pyspark")
            expect(json_body[:environmentName]).to eq "spark-feature-pipeline"
          end
          it "check default spark job image" do
            get_default_job_configuration(@project[:id], "spark")
            expect(json_body[:environmentName]).to eq "spark-feature-pipeline"
          end
          it "check default flink job image" do
            get_default_job_configuration(@project[:id], "flink")
            expect(json_body[:environmentName]).to eq "spark-feature-pipeline"
          end
          it "update job image" do
            job_name="custom_job1"
            job_conf = get_python_default_config(@project, job_name, "py")
            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/jobs/#{job_name}", job_conf
            expect(json_body[:config][:environmentName]).to eq "python-feature-pipeline"

            job_conf["environmentName"]="torch-training-pipeline"
            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/jobs/#{job_name}", job_conf

            get_job(@project[:id], job_name)
            expect(json_body[:config][:environmentName]).to eq "torch-training-pipeline"

            get_job(@project[:id], job_name, query: "?expand=environments")
            expect(json_body[:config][:environmentName]).to eq "torch-training-pipeline"
            expect(json_body[:environment][:name]).to eq "torch-training-pipeline"
          end
          it "should fail when configuring to non-existent environment" do
            job_name="custom_job2"
            job_conf = get_python_default_config(@project, job_name, "py")
            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/jobs/#{job_name}", job_conf
            expect(json_body[:config][:environmentName]).to eq "python-feature-pipeline"

            job_conf["environmentName"]="i_do_not_exist"
            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/jobs/#{job_name}", job_conf
            expect_status_details(404)
          end
          it "should fail when configuring to not supported environment" do
            job_name="custom_job3"
            job_conf = get_python_default_config(@project, job_name, "py")
            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/jobs/#{job_name}", job_conf
            expect(json_body[:config][:environmentName]).to eq "python-feature-pipeline"

            job_conf["environmentName"]="minimal-inference-pipeline"
            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/jobs/#{job_name}", job_conf
            expect_status_details(400, error_code: 300021)
          end
          it "create job with not supported environment" do
            job_name="custom_job4"
            job_conf = get_python_default_config(@project, job_name, "py")
            job_conf["environmentName"]="minimal-inference-pipeline"
            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/jobs/#{job_name}", job_conf
            expect_status_details(400, error_code: 300021)
          end
          it "should fail as module not installed" do
            job_name="test_env_job"
            create_env_check_job(@project, job_name, "tensorflow", "test_env_job.py", "python-feature-pipeline")

            get_job(@project[:id], job_name, query: "?expand=environments")
            #expect env name set
            expect(json_body[:config][:environmentName]).to eq "python-feature-pipeline"
            expect(json_body[:environment][:name]).to eq "python-feature-pipeline"

            #start execution
            start_execution(@project[:id], job_name)
            execution_id = json_body[:id]
            expect(json_body[:state]).to eq "INITIALIZING"
            expect(json_body[:finalStatus]).to eq "UNDEFINED"

            #get execution
            get_execution(@project[:id], job_name, json_body[:id])
            expect(json_body[:id]).to eq(execution_id)
            #wait till it's finished and start second execution
            wait_for_execution_completed(@project[:id], job_name, json_body[:id], "FAILED", expected_final_status: "FAILED")
          end
          it "should run in python-feature-pipeline" do
            job_name="test_env_job"
            create_env_check_job(@project, job_name, "hopsworks", "test_env_job.py", "python-feature-pipeline")

            get_job(@project[:id], job_name, query: "?expand=environments")
            #expect env name set
            expect(json_body[:config][:environmentName]).to eq "python-feature-pipeline"
            expect(json_body[:environment][:name]).to eq "python-feature-pipeline"

            #start execution
            start_execution(@project[:id], job_name)
            execution_id = json_body[:id]
            expect(json_body[:state]).to eq "INITIALIZING"
            expect(json_body[:finalStatus]).to eq "UNDEFINED"

            #get execution
            get_execution(@project[:id], job_name, json_body[:id])
            expect(json_body[:id]).to eq(execution_id)
            #wait till it's finished and start second execution
            wait_for_execution_completed(@project[:id], job_name, json_body[:id], "FINISHED", expected_final_status: "SUCCEEDED")
          end
          it "should run in torch-training-pipeline" do
            job_name="test_env_job"
            create_env_check_job(@project, job_name, "torch", "test_env_job.py", "torch-training-pipeline")

            get_job(@project[:id], job_name, query: "?expand=environments")
            #expect env name set
            expect(json_body[:config][:environmentName]).to eq "torch-training-pipeline"
            expect(json_body[:environment][:name]).to eq "torch-training-pipeline"

            #start execution
            start_execution(@project[:id], job_name)
            execution_id = json_body[:id]
            expect(json_body[:state]).to eq "INITIALIZING"
            expect(json_body[:finalStatus]).to eq "UNDEFINED"

            #get execution
            get_execution(@project[:id], job_name, json_body[:id])
            expect(json_body[:id]).to eq(execution_id)
            #wait till it's finished and start second execution
            wait_for_execution_completed(@project[:id], job_name, json_body[:id], "FINISHED", expected_final_status: "SUCCEEDED")
          end
          it "should run in tensorflow-training-pipeline" do
            job_name="test_env_job"
            create_env_check_job(@project, job_name, "tensorflow", "test_env_job.py", "tensorflow-training-pipeline")

            get_job(@project[:id], job_name, query: "?expand=environments")
            #expect env name set
            expect(json_body[:config][:environmentName]).to eq "tensorflow-training-pipeline"
            expect(json_body[:environment][:name]).to eq "tensorflow-training-pipeline"

            #start execution
            start_execution(@project[:id], job_name)
            execution_id = json_body[:id]
            expect(json_body[:state]).to eq "INITIALIZING"
            expect(json_body[:finalStatus]).to eq "UNDEFINED"

            #get execution
            get_execution(@project[:id], job_name, json_body[:id])
            expect(json_body[:id]).to eq(execution_id)
            #wait till it's finished and start second execution
            wait_for_execution_completed(@project[:id], job_name, json_body[:id], "FINISHED", expected_final_status: "SUCCEEDED")
          end
          it "should run in spark-feature-pipeline" do
            job_name="test_env_job"
            create_env_check_job(@project, job_name, "sparkmagic", "test_env_job.py", "spark-feature-pipeline")

            get_job(@project[:id], job_name, query: "?expand=environments")
            #expect env name set
            expect(json_body[:config][:environmentName]).to eq "spark-feature-pipeline"
            expect(json_body[:environment][:name]).to eq "spark-feature-pipeline"

            #start execution
            start_execution(@project[:id], job_name)
            execution_id = json_body[:id]
            expect(json_body[:state]).to eq "INITIALIZING"
            expect(json_body[:finalStatus]).to eq "UNDEFINED"

            #get execution
            get_execution(@project[:id], job_name, json_body[:id])
            expect(json_body[:id]).to eq(execution_id)
            #wait till it's finished and start second execution
            wait_for_execution_completed(@project[:id], job_name, json_body[:id], "FINISHED", expected_final_status: "SUCCEEDED")
          end
          it "should fail when trying to remove environment in use" do
            job_name="fk_removal_job"
            create_environment(@project[:id], "job_env", "python-feature-pipeline", "some description")
            wait_for_running_command(@project[:id])

            create_python_job(@project, job_name, "py")

            get_job(@project[:id], job_name, query: "?expand=environments")
            job = json_body
            job[:config][:environmentName] = "job_env"

            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/jobs/#{job_name}", job[:config]
            expect_status_details(200)

            # Should fail as environment is in use
            delete_env(@project[:id], "job_env", expected_status: 400, error_code: 300022)
            expect(json_body[:usrMsg]).to include job_name

            delete_job(@project[:id], job[:name])
            expect_status_details(204)

            delete_env(@project[:id], "job_env")
            expect_status_details(204)
          end
        end
        context 'for kserve deployment' do
          before :all do
            if ENV['OS'] == "ubuntu"
              skip "These tests do not run on ubuntu"
            end
            setVar('enable_data_science_profile', "true")
            with_valid_project
            copy_iris_files(@project[:projectname], @user[:username])
            copy_mnist_files(@project[:projectname], @user[:username])
          end
          it "should have default misc environment set" do
            # Create kafka topic
            json, topic_name = add_topic(@project[:id], INFERENCE_SCHEMA_NAME, 4)

            # Create serving
            name = "testdep1"
            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/serving/", parse_serving_json(
                {name: name,
                modelPath: "/Projects/#{@project[:projectname]}/Models/irisflowerclassifier",
                modelVersion: 1,
                kafkaTopicDTO: {
                  name: topic_name
                },
                inferenceLogging: "ALL",
                modelServer: "PYTHON",
                modelFramework: "SKLEARN",
                servingTool: "KSERVE",
                requestedInstances: 1
                })
            expect_status_details(201)

            expect(json_body[:environmentDTO][:name]).to eq "misc-inference-pipeline"
          end
          it "should have default sklearn environment set" do
            # Create kafka topic
            json, topic_name = add_topic(@project[:id], INFERENCE_SCHEMA_NAME, 4)

            # Create serving
            name = "testdep2"
            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/serving/", parse_serving_json(
                {name: name,
                modelPath: "/Projects/#{@project[:projectname]}/Models/irisflowerclassifier",
                modelVersion: 1,
                kafkaTopicDTO: {
                  name: topic_name
                },
                inferenceLogging: "ALL",
                modelServer: "PYTHON",
                modelFramework: "TORCH",
                servingTool: "KSERVE",
                requestedInstances: 1,
                predictor: "/Projects/#{@project[:projectname]}/Models/irisflowerclassifier/1/#{SKLEARN_SCRIPT_FILE_NAME}"
                })
            expect_status_details(201)

            expect(json_body[:environmentDTO][:name]).to eq "torch-inference-pipeline"
          end
          it "should have default torch environment set" do
            # Create kafka topic
            json, topic_name = add_topic(@project[:id], INFERENCE_SCHEMA_NAME, 4)

            # Create serving
            name = "testdep3"
            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/serving/", parse_serving_json(
                {name: name,
                modelPath: "/Projects/#{@project[:projectname]}/Models/irisflowerclassifier",
                modelVersion: 1,
                kafkaTopicDTO: {
                  name: topic_name
                },
                inferenceLogging: "ALL",
                modelServer: "PYTHON",
                modelFramework: "TORCH",
                servingTool: "KSERVE",
                requestedInstances: 1,
                predictor: "/Projects/#{@project[:projectname]}/Models/irisflowerclassifier/1/#{SKLEARN_SCRIPT_FILE_NAME}"
                })
            expect_status_details(201)

            expect(json_body[:environmentDTO][:name]).to eq "torch-inference-pipeline"
          end
          it "should have default python environment set" do
            # Create kafka topic
            json, topic_name = add_topic(@project[:id], INFERENCE_SCHEMA_NAME, 4)

            # Create serving
            name = "testdep4"
            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/serving/", parse_serving_json(
                {name: name,
                modelPath: "/Projects/#{@project[:projectname]}/Models/irisflowerclassifier",
                modelVersion: 1,
                kafkaTopicDTO: {
                  name: topic_name
                },
                inferenceLogging: "ALL",
                modelServer: "PYTHON",
                modelFramework: "PYTHON",
                servingTool: "KSERVE",
                requestedInstances: 1,
                predictor: "/Projects/#{@project[:projectname]}/Models/irisflowerclassifier/1/#{SKLEARN_SCRIPT_FILE_NAME}"
                })
            expect_status_details(201)

            expect(json_body[:environmentDTO][:name]).to eq "misc-inference-pipeline"
          end
          it "should be able to update deployment environment" do
            # Create kafka topic
            json, topic_name = add_topic(@project[:id], INFERENCE_SCHEMA_NAME, 4)

            # Create serving
            name = "testdep5"
            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/serving/", parse_serving_json(
                {name: name,
                modelPath: "/Projects/#{@project[:projectname]}/Models/irisflowerclassifier",
                modelVersion: 1,
                kafkaTopicDTO: {
                  name: topic_name
                },
                inferenceLogging: "ALL",
                modelServer: "PYTHON",
                modelFramework: "SKLEARN",
                servingTool: "KSERVE",
                requestedInstances: 1
                })
            expect_status_details(201)

            config = json_body

            create_environment(@project[:id], "minimal-inference-pipeline-test", "minimal-inference-pipeline", "some description")
            wait_for_running_command(@project[:id])

            config[:environmentDTO][:name] = "minimal-inference-pipeline-test"
            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/serving/", config
            expect_status_details(201)
            expect(json_body[:environmentDTO][:name]).to eq "minimal-inference-pipeline-test"

            serving = get_serving(@project, name)
            expect(serving["environmentDTO"]["name"]).to eq "minimal-inference-pipeline-test"
          end
          it "should fail when creating with non-existent environment" do
            # Create kafka topic
            json, topic_name = add_topic(@project[:id], INFERENCE_SCHEMA_NAME, 4)

            # Create serving
            name = "testdep6"
            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/serving/", parse_serving_json(
                {name: name,
                modelPath: "/Projects/#{@project[:projectname]}/Models/irisflowerclassifier",
                modelVersion: 1,
                kafkaTopicDTO: {
                  name: topic_name
                },
                inferenceLogging: "ALL",
                modelServer: "PYTHON",
                modelFramework: "SKLEARN",
                servingTool: "KSERVE",
                requestedInstances: 1,
                environmentDTO: {
                  name: "i_do_not_exist"
                }
                })
            expect_status_details(404)
          end
          it "should fail when updating to non-existent environment" do
            # Create kafka topic
            json, topic_name = add_topic(@project[:id], INFERENCE_SCHEMA_NAME, 4)

            # Create serving
            name = "testdep7"
            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/serving/", parse_serving_json(
                {name: name,
                modelPath: "/Projects/#{@project[:projectname]}/Models/irisflowerclassifier",
                modelVersion: 1,
                kafkaTopicDTO: {
                  name: topic_name
                },
                inferenceLogging: "ALL",
                modelServer: "PYTHON",
                modelFramework: "SKLEARN",
                servingTool: "KSERVE",
                requestedInstances: 1
                })
            expect_status_details(201)

            config = json_body

            config[:environmentDTO][:name] = "i_do_not_exist"
            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/serving/", config
            expect_status_details(404)
          end
          it "should fail when trying to remove environment in use" do
            create_environment(@project[:id], "deployment_env", "minimal-inference-pipeline", "some description")
            wait_for_running_command(@project[:id])

            # Create kafka topic
            json, topic_name = add_topic(@project[:id], INFERENCE_SCHEMA_NAME, 4)

            # Create serving
            name = "testdep8"
            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/serving/", parse_serving_json(
                {name: name,
                modelPath: "/Projects/#{@project[:projectname]}/Models/irisflowerclassifier",
                modelVersion: 1,
                kafkaTopicDTO: {
                  name: topic_name
                },
                inferenceLogging: "ALL",
                modelServer: "PYTHON",
                modelFramework: "SKLEARN",
                servingTool: "KSERVE",
                requestedInstances: 1,
                environmentDTO: {
                  name: "deployment_env"
                }
                })
            expect_status_details(201)
            serving_id = json_body[:id]

            delete_env(@project[:id], "deployment_env", expected_status: 400, error_code: 300022)
            expect(json_body[:usrMsg]).to include name

            delete_serving(@project[:id], serving_id)

            delete_env(@project[:id], "deployment_env")
            expect_status_details(204)
          end
          it "should fail when updating to not supported environment" do
            # Create kafka topic
            json, topic_name = add_topic(@project[:id], INFERENCE_SCHEMA_NAME, 4)

            # Create serving
            name = "testdep9"
            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/serving/", parse_serving_json(
                {name: name,
                modelPath: "/Projects/#{@project[:projectname]}/Models/irisflowerclassifier",
                modelVersion: 1,
                kafkaTopicDTO: {
                  name: topic_name
                },
                inferenceLogging: "ALL",
                modelServer: "PYTHON",
                modelFramework: "SKLEARN",
                servingTool: "KSERVE",
                requestedInstances: 1
                })
            expect_status_details(201)

            config = json_body

            config[:environmentDTO][:name] = "python-feature-pipeline"
            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/serving/", config
            expect_status_details(400, error_code: 300021)
          end
          it "should fail when creating with not supported environment" do
            # Create kafka topic
            json, topic_name = add_topic(@project[:id], INFERENCE_SCHEMA_NAME, 4)

            # Create serving
            name = "testdep10"
            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/serving/", parse_serving_json(
                {name: name,
                modelPath: "/Projects/#{@project[:projectname]}/Models/irisflowerclassifier",
                modelVersion: 1,
                kafkaTopicDTO: {
                  name: topic_name
                },
                inferenceLogging: "ALL",
                modelServer: "PYTHON",
                modelFramework: "SKLEARN",
                servingTool: "KSERVE",
                requestedInstances: 1,
                environmentDTO: {
                  name: "tensorflow-training-pipeline"
                }
                })
            expect_status_details(400, error_code: 300021)
          end
          it "should create a kserve serving with tensorflow serving with no environment set" do
            name = "testmodel16"
            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/serving/", parse_serving_json(
                {name: name,
                 modelPath: "/Projects/#{@project[:projectname]}/Models/mnist",
                 batchingConfiguration: {
                   batchingEnabled: false
                 },
                 modelVersion: 1,
                 modelServer: "TENSORFLOW_SERVING",
                 modelFramework: "TENSORFLOW",
                 servingTool: "KSERVE",
                 requestedInstances: 1
                })
            expect_status_details(201)

            config = json_body
            expect(config[:environmentDTO]).to be_nil
          end
          it 'should create a non-kserve serving that defaults to correct environment' do
            name = "testModelREST"
            put "#{ENV['HOPSWORKS_API']}/project/#{@project[:id]}/serving/", parse_serving_json(
                {name: name,
                 modelPath: "/Projects/#{@project[:projectname]}/Models/irisflowerclassifier",
                 modelVersion: 1,
                 predictor: "/Projects/#{@project[:projectname]}/Models/irisflowerclassifier/1/#{SKLEARN_SCRIPT_FILE_NAME}",
                 modelServer: "PYTHON",
                 modelFramework: "TORCH",
                 servingTool: "DEFAULT",
                 requestedInstances: 1
                })
            expect_status_details(201)

            config = json_body
            expect(config[:environmentDTO][:name]).to eq ('torch-inference-pipeline')
          end
        end
      end
    end
  end
end
