# This file is part of Hopsworks
# Copyright (C) 2022, Logical Clocks AB. All rights reserved
#
# Hopsworks is free software: you can redistribute it and/or modify it under the terms of
# the GNU Affero General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Hopsworks is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with this program.
# If not, see <https://www.gnu.org/licenses/>.

require 'json'

describe "On #{ENV['OS']}" do
  after(:all) { clean_all_test_projects(spec: "featureviewtransformation") }

  describe "feature view transformation" do
    describe "internal" do
      context 'with valid project, featurestore service enabled' do
        before :all do
          with_valid_project
        end

        it "should be able to attach transformation function." do
          featurestore_id = get_featurestore_id(@project.id)
          featuregroup_suffix = short_random_id
          query = make_sample_query(@project, featurestore_id, featuregroup_suffix: featuregroup_suffix)

          # create transformation function
          json_result = register_transformation_fn(@project.id, featurestore_id)
          transformation_function_feature_a = JSON.parse(json_result)
          expect_status_details(200)

          feature_schema = [
            {type: "INT", name: "a_testfeature1", featureGroupFeatureName: "a_testfeature1", label: false},
            {type: "INT", name: "b_b_testfeature1", featureGroupFeatureName: "b_testfeature1", label: false}
          ]
          transformation_function_feature_a["hopsworksUdf"]["transformationFeatures"] = ["a_testfeature1"]
          transformation_function_feature_a["hopsworksUdf"]["outputTypes"] = ["double"]

          transformation_feature_b = {
            id: nil,
            version: 1,
            featurestoreId: 11,
            hopsworksUdf: {
              sourceCode: "\\n@hopworks.udf(int)\\ndef plus_two(data1 : pd.Series):\\n    return data1 + 2",
              name: "plus_two",
              outputTypes: ["bigint"],
              transformationFeatures: ["b_b_testfeature1"],
              transformationFunctionArgumentNames: ["data1"],
            }
          }

          json_result = create_feature_view(@project.id, featurestore_id, query, features: feature_schema, transformations: [transformation_function_feature_a, transformation_feature_b])
          parsed_json = JSON.parse(json_result)
          expect_status_details(201)
          feature_view_name = parsed_json["name"]
          feature_view_version = parsed_json["version"]

          # endpoint to attach transformation functions
          json_result = get "#{ENV['HOPSWORKS_API']}/project/#{@project.id}/featurestores/#{featurestore_id}/featureview/#{feature_view_name}/version/#{feature_view_version}/transformation"
          parsed_json = JSON.parse(json_result)
          feature_transformation = parsed_json["items"].select{ |f| f["hopsworksUdf"]["name"] == "plus_one"}.first
          expect(feature_transformation["hopsworksUdf"]["name"]).to eql("plus_one")
          expect(feature_transformation["hopsworksUdf"]["outputTypes"]).to eql(["double"])
          expect(feature_transformation["hopsworksUdf"]["transformationFeatures"]).to eql(["a_testfeature1"])
          expect(feature_transformation["hopsworksUdf"]["transformationFunctionArgumentNames"]).to eql(["data1"])
          expect(feature_transformation["hopsworksUdf"]["droppedArgumentNames"]).to eql(["data1"])
          expect(feature_transformation["hopsworksUdf"]["statisticsArgumentNames"]).to eql(nil)

          feature_transformation = parsed_json["items"].select{ |f| f["hopsworksUdf"]["name"] == "plus_two"}.first
          expect(feature_transformation["hopsworksUdf"]["name"]).to eql("plus_two")
          expect(feature_transformation["hopsworksUdf"]["outputTypes"]).to eql(["bigint"])
          expect(feature_transformation["hopsworksUdf"]["transformationFeatures"]).to eql(["b_b_testfeature1"])
          expect(feature_transformation["hopsworksUdf"]["transformationFunctionArgumentNames"]).to eql(["data1"])
          expect(feature_transformation["hopsworksUdf"]["droppedArgumentNames"]).to eql(nil)
          expect(feature_transformation["hopsworksUdf"]["statisticsArgumentNames"]).to eql(nil)

        end

        it "should be able to create feature view using feature groups with on-demand transformations" do
          featurestore_id = get_featurestore_id(@project.id)
          query = make_sample_query_on_demand(@project, featurestore_id)

          json_result = create_feature_view(@project.id, featurestore_id, query)
          parsed_json = JSON.parse(json_result)
          expect_status_details(201)

          # endpoint to attach transformation functions
          parsed_json = JSON.parse(json_result)

          on_demand_feature = parsed_json["features"].select{ |f| f["name"] == "a_on_demand_feature"}.first
          feature_transformation = on_demand_feature["transformationFunction"]
          expect(feature_transformation["hopsworksUdf"]["name"]).to eql("a_on_demand_feature")
          expect(feature_transformation["hopsworksUdf"]["outputTypes"]).to eql(["bigint"])
          expect(feature_transformation["hopsworksUdf"]["transformationFeatures"]).to eql(["a_testfeature1"])
          expect(feature_transformation["hopsworksUdf"]["transformationFunctionArgumentNames"]).to eql(["data1"])
          expect(feature_transformation["hopsworksUdf"]["droppedArgumentNames"]).to eql(["data1"])
          expect(feature_transformation["hopsworksUdf"]["statisticsArgumentNames"]).to eql(nil)

          on_demand_feature = parsed_json["features"].select{ |f| f["name"] == "fgb_b_on_demand_feature"}.first
          feature_transformation = on_demand_feature["transformationFunction"]
          expect(feature_transformation["hopsworksUdf"]["name"]).to eql("b_on_demand_feature")
          expect(feature_transformation["hopsworksUdf"]["outputTypes"]).to eql(["double"])
          expect(feature_transformation["hopsworksUdf"]["transformationFeatures"]).to eql(["b_testfeature1"])
          expect(feature_transformation["hopsworksUdf"]["transformationFunctionArgumentNames"]).to eql(["data1"])
          expect(feature_transformation["hopsworksUdf"]["droppedArgumentNames"]).to eql(nil)
          expect(feature_transformation["hopsworksUdf"]["statisticsArgumentNames"]).to eql(nil)
        end

        it "should be able to create feature view using feature groups with on-demand transformations and attach model dependent transformations" do
          featurestore_id = get_featurestore_id(@project.id)
          query = make_sample_query_on_demand(@project, featurestore_id)

          # create transformation function

          feature_schema = [
            { type: "INT", name: "a_testfeature", featureGroupFeatureName: "a_testfeature", label: false},
            { type: "INT", name: "a_on_demand_feature", featureGroupFeatureName: "a_on_demand_feature", label: false },
            { type: "INT", name: "fgb_b_on_demand_feature", featureGroupFeatureName: "b_on_demand_feature", label: false },
            {type: "INT", name: "fgb_a_testfeature", featureGroupFeatureName: "a_testfeature", label: false}
          ]

          transformation_feature = {
            id: nil,
            version: 1,
            featurestoreId: 11,
            hopsworksUdf: {
              sourceCode: "\\n@hopworks.udf(int)\\ndef plus_two(data1 : pd.Series):\\n    return data1 + 2",
              name: "plus_two",
              outputTypes: ["bigint"],
              transformationFeatures: ["a_testfeature"],
              transformationFunctionArgumentNames: ["data1"],
            }
          }

          json_result = create_feature_view(@project.id, featurestore_id, query, features: feature_schema, transformations: [transformation_feature])
          parsed_json = JSON.parse(json_result)
          expect_status_details(201)

          on_demand_feature = parsed_json["features"].select{ |f| f["name"] == "a_on_demand_feature"}.first
          feature_transformation = on_demand_feature["transformationFunction"]
          expect(feature_transformation["hopsworksUdf"]["name"]).to eql("a_on_demand_feature")
          expect(feature_transformation["hopsworksUdf"]["outputTypes"]).to eql(["bigint"])
          expect(feature_transformation["hopsworksUdf"]["transformationFeatures"]).to eql(["a_testfeature1"])
          expect(feature_transformation["hopsworksUdf"]["transformationFunctionArgumentNames"]).to eql(["data1"])
          expect(feature_transformation["hopsworksUdf"]["droppedArgumentNames"]).to eql(["data1"])
          expect(feature_transformation["hopsworksUdf"]["statisticsArgumentNames"]).to eql(nil)

          on_demand_feature = parsed_json["features"].select{ |f| f["name"] == "fgb_b_on_demand_feature"}.first
          feature_transformation = on_demand_feature["transformationFunction"]
          expect(feature_transformation["hopsworksUdf"]["name"]).to eql("b_on_demand_feature")
          expect(feature_transformation["hopsworksUdf"]["outputTypes"]).to eql(["double"])
          expect(feature_transformation["hopsworksUdf"]["transformationFeatures"]).to eql(["b_testfeature1"])
          expect(feature_transformation["hopsworksUdf"]["transformationFunctionArgumentNames"]).to eql(["data1"])
          expect(feature_transformation["hopsworksUdf"]["droppedArgumentNames"]).to eql(nil)
          expect(feature_transformation["hopsworksUdf"]["statisticsArgumentNames"]).to eql(nil)

          # endpoint to attach transformation functions
          feature_view_name = parsed_json["name"]
          feature_view_version = parsed_json["version"]
          json_result = get "#{ENV['HOPSWORKS_API']}/project/#{@project.id}/featurestores/#{featurestore_id}/featureview/#{feature_view_name}/version/#{feature_view_version}/transformation"
          parsed_json = JSON.parse(json_result)

          feature_transformation = parsed_json["items"].select{ |f| f["hopsworksUdf"]["name"] == "plus_two"}.first
          expect(feature_transformation["hopsworksUdf"]["name"]).to eql("plus_two")
          expect(feature_transformation["hopsworksUdf"]["outputTypes"]).to eql(["bigint"])
          expect(feature_transformation["hopsworksUdf"]["transformationFeatures"]).to eql(["a_testfeature"])
          expect(feature_transformation["hopsworksUdf"]["transformationFunctionArgumentNames"]).to eql(["data1"])
          expect(feature_transformation["hopsworksUdf"]["droppedArgumentNames"]).to eql(nil)
          expect(feature_transformation["hopsworksUdf"]["statisticsArgumentNames"]).to eql(nil)
        end

      end
    end
  end
end