# This file is part of Hopsworks
# Copyright (C) 2020, Logical Clocks AB. All rights reserved
#
# Hopsworks is free software: you can redistribute it and/or modify it under the terms of
# the GNU Affero General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Hopsworks is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with this program.
# If not, see <https://www.gnu.org/licenses/>.

require 'json'

describe "On #{ENV['OS']}" do
  after(:all) {clean_all_test_projects(spec: "preparedstatements")}

  describe "prepared statements" do

    describe "list" do
      context 'with valid project, featurestore service enabled' do
        before :all do
          with_valid_project
        end

		it "should be able to get a feature view serving vector" do
          # create first feature group
          featurestore_id = get_featurestore_id(@project.id)
          project_name = @project.projectname
          features = ['a', 'b', 'c', 'd'].map do  |feat_name|
              {type: "INT", name: feat_name}
          end
          features[0]['primary'] = true
          json_result, fg_name = create_cached_featuregroup(@project.id, featurestore_id, features: features,
                                                            featuregroup_name: "test_fg_a_#{short_random_id}", online:true)
          parsed_json = JSON.parse(json_result)
          expect_status_details(201)
          fg_id = parsed_json["id"]
          fg_type = parsed_json["type"]
          # create queryDTO object
          query = {
            leftFeatureGroup: {
              id: fg_id,
              type: fg_type
            },
            leftFeatures: ['d', 'c', 'a', 'b'].map do |feat_name|
              {name: feat_name}
            end,
            joins: []
          }

          json_result = create_feature_view(@project.id, featurestore_id, query)
          parsed_json = JSON.parse(json_result)
          expect_status_details(201)

          feature_view_name = parsed_json["name"]
          feature_view_version = parsed_json["version"]
          json_result = get "#{ENV['HOPSWORKS_API']}/project/#{@project.id}/featurestores/#{featurestore_id}/featureview/#{feature_view_name}/version/#{feature_view_version}/preparedstatement"
          expect_status_details(200)

          parsed_json = JSON.parse(json_result)
          expect(parsed_json["items"].first["preparedStatementParameters"].first["index"]).to eql(1)
          expect(parsed_json["items"].first["preparedStatementParameters"].first["name"]).to eql("a")
          expect(parsed_json["items"].first["queryOnline"]).to eql("SELECT `fg0`.`d` AS `d`, `fg0`.`c` AS `c`, `fg0`.`a` AS `a`, `fg0`.`b` AS `b`\n"+
                                                                   "FROM `#{project_name.downcase}`.`#{fg_name}_1` AS `fg0`\n" +
                                                                   "WHERE `fg0`.`a` = ?")
        end

        it "should be able to get a feature view serving vector in correct order" do
          # create first feature group
          featurestore_id = get_featurestore_id(@project.id)
          project_name = @project.projectname
          features = [
              {type: "INT", name: "a_testfeature", primary: true},
              {type: "INT", name: "a_testfeature1"},
          ]
          json_result, fg_name = create_cached_featuregroup(@project.id, featurestore_id, features: features, featuregroup_name: "test_fg_a_#{short_random_id}", online:true)
          parsed_json = JSON.parse(json_result)
          expect_status_details(201)
          fg_id = parsed_json["id"]
          fg_type = parsed_json["type"]
          # create second feature group
          features = [
              {type: "INT", name: "a_testfeature", primary: true},
              {type: "INT", name: "b_testfeature1"},
          ]
          json_result_b, fg_name_b = create_cached_featuregroup(@project.id, featurestore_id, features: features, featuregroup_name: "test_fg_b_#{short_random_id}", online:true)
          parsed_json_b = JSON.parse(json_result_b)
          expect_status_details(201)
          fg_id_b = parsed_json_b["id"]
          fg_b_type = parsed_json_b["type"]
          # create queryDTO object
          query = {
              leftFeatureGroup: {
                  id: fg_id,
                  type: fg_type
              },
              leftFeatures: [{name: 'a_testfeature1'}],
              joins: [{
                       query: {
                           leftFeatureGroup: {
                               id: fg_id_b,
                               type: fg_b_type
                           },
                           leftFeatures: [{name: 'b_testfeature1'}]
                       }
                  }
              ]
          }

          json_result = create_feature_view(@project.id, featurestore_id, query)
          parsed_json = JSON.parse(json_result)
          expect_status_details(201)

          feature_view_name = parsed_json["name"]
          feature_view_version = parsed_json["version"]
          json_result = get "#{ENV['HOPSWORKS_API']}/project/#{@project.id}/featurestores/#{featurestore_id}/featureview/#{feature_view_name}/version/#{feature_view_version}/preparedstatement"
          expect_status_details(200)

          parsed_json = JSON.parse(json_result)
          expect(parsed_json["items"].first["preparedStatementParameters"].first["index"]).to eql(1)
          expect(parsed_json["items"].first["preparedStatementParameters"].first["name"]).to eql("a_testfeature")
          expect(parsed_json["items"].first["queryOnline"]).to eql("SELECT `fg0`.`a_testfeature1` AS `a_testfeature1`\nFROM `#{project_name.downcase}`.`#{fg_name}_1` AS `fg0`\nWHERE `fg0`.`a_testfeature` = ?")
          expect(parsed_json["items"].second["preparedStatementParameters"].first["index"]).to eql(1)
          expect(parsed_json["items"].second["preparedStatementParameters"].first["name"]).to eql("a_testfeature")
          expect(parsed_json["items"].second["queryOnline"]).to eql("SELECT `fg0`.`b_testfeature1` AS `b_testfeature1`\nFROM `#{project_name.downcase}`.`#{fg_name_b}_1` AS `fg0`\nWHERE `fg0`.`a_testfeature` = ?")
        end

        it "should be able to get a feature view serving vector in correct order and remove feature group with only primary key and label" do
          # create first feature group
          featurestore_id = get_featurestore_id(@project.id)
          project_name = @project.projectname
          features = [
              {type: "INT", name: "a_testfeature", primary: true},
              {type: "INT", name: "a_testfeature1"},
          ]
          json_result, fg_name = create_cached_featuregroup(@project.id, featurestore_id, features: features, featuregroup_name: "test_fg_a_#{short_random_id}", online:true)
          parsed_json = JSON.parse(json_result)
          expect_status_details(201)
          fg_id = parsed_json["id"]
          fg_type = parsed_json["type"]
          # create second feature group
          features = [
              {type: "INT", name: "a_testfeature", primary: true},
              {type: "INT", name: "b_testfeature1"},
          ]
          json_result_b, fg_name_b = create_cached_featuregroup(@project.id, featurestore_id, features: features, featuregroup_name: "test_fg_b_#{short_random_id}", online:true)
          parsed_json_b = JSON.parse(json_result_b)
          expect_status_details(201)
          fg_id_b = parsed_json_b["id"]
          fg_b_type = parsed_json_b["type"]
          # create third feature group
          features = [
              {type: "INT", name: "a_testfeature", primary: true},
              {type: "INT", name: "c_testfeature1",label: true},
          ]
          json_result_c, fg_name_c = create_cached_featuregroup(@project.id, featurestore_id, features: features, featuregroup_name: "test_fg_c_#{short_random_id}", online:true)
          parsed_json_c = JSON.parse(json_result_c)
          expect_status_details(201)
          fg_id_c = parsed_json_c["id"]
          fg_c_type = parsed_json_c["type"]
          # create queryDTO object
          query = {
              leftFeatureGroup: {
                  id: fg_id,
                  type: fg_type
              },
              leftFeatures: [{name: 'a_testfeature1'}],
              joins: [{
                          query: {
                              leftFeatureGroup: {
                                  id: fg_id_c,
                                  type: fg_c_type
                              },
                              leftFeatures: [{name: 'c_testfeature1'}]
                          },
                      },
                      {
                          query: {
                              leftFeatureGroup: {
                                  id: fg_id_b,
                                  type: fg_b_type
                              },
                              leftFeatures: [{name: 'b_testfeature1'}]
                          },
                      }
              ]
          }

          feature_schema = [
              {type: "INT", name: "a_testfeature1", label: false},
              {type: "INT", name: "b_testfeature1", label: false},
              {type: "INT", name: "c_testfeature1", label: true}
          ]

          json_result = create_feature_view(@project.id, featurestore_id, query, features: feature_schema)
          parsed_json = JSON.parse(json_result)
          expect_status_details(201)

          feature_view_name = parsed_json["name"]
          feature_view_version = parsed_json["version"]
          json_result = get "#{ENV['HOPSWORKS_API']}/project/#{@project.id}/featurestores/#{featurestore_id}/featureview/#{feature_view_name}/version/#{feature_view_version}/preparedstatement?batch=false"
          expect_status_details(200)

          parsed_json = JSON.parse(json_result)
          expect(parsed_json["items"].length).to eql(2)
          expect(parsed_json["items"].first["preparedStatementParameters"].first["index"]).to eql(1)
          expect(parsed_json["items"].first["preparedStatementParameters"].first["name"]).to eql("a_testfeature")
          expect(parsed_json["items"].first["queryOnline"]).to eql("SELECT `fg0`.`a_testfeature1` AS `a_testfeature1`\nFROM `#{project_name.downcase}`.`#{fg_name}_1` AS `fg0`\nWHERE `fg0`.`a_testfeature` = ?")
          expect(parsed_json["items"].second["preparedStatementParameters"].first["index"]).to eql(1)
          expect(parsed_json["items"].second["preparedStatementParameters"].first["name"]).to eql("a_testfeature")
          expect(parsed_json["items"].second["queryOnline"]).to eql("SELECT `fg0`.`b_testfeature1` AS `b_testfeature1`\nFROM `#{project_name.downcase}`.`#{fg_name_b}_1` AS `fg0`\nWHERE `fg0`.`a_testfeature` = ?")
        end

        it "it should add primary keys to the prepared statement in batch mode" do
          # create first feature group
          featurestore_id = get_featurestore_id(@project.id)
          project_name = @project.projectname
          features = [
            {type: "INT", name: "a_testfeature", primary: true},
            {type: "INT", name: "a_testfeature1"},
          ]
          json_result, fg_name = create_cached_featuregroup(@project.id, featurestore_id, features: features, featuregroup_name: "test_fg_a_#{short_random_id}", online:true)
          parsed_json = JSON.parse(json_result)
          expect_status_details(201)
          fg_id = parsed_json["id"]
          fg_type = parsed_json["type"]
          # create second feature group
          features = [
            {type: "INT", name: "b_testfeature", primary: true},
            {type: "INT", name: "b_testfeature1"},
          ]
          json_result_b, fg_name_b = create_cached_featuregroup(@project.id, featurestore_id, features: features, featuregroup_name: "test_fg_b_#{short_random_id}", online:true)
          parsed_json_b = JSON.parse(json_result_b)
          expect_status_details(201)
          fg_id_b = parsed_json_b["id"]
          fg_b_type = parsed_json_b["type"]
          # create queryDTO object
          query = {
            leftFeatureGroup: {
              id: fg_id,
              type: fg_type
            },
            leftFeatures: [{name: 'a_testfeature1'}],
            joins: [{
                      query: {
                        leftFeatureGroup: {
                          id: fg_id_b,
                          type: fg_b_type
                        },
                        leftFeatures: [{name: 'b_testfeature1'}],
                      },
                      leftOn: [{name: 'a_testfeature'}],
                      rightOn: [{name: 'b_testfeature'}],
                      prefix: 'yolo_'
                   }]
          }

          feature_schema = [
            {type: "INT", name: "a_testfeature1", label: false},
            {type: "INT", name: "b_testfeature1", label: false},
          ]

          json_result = create_feature_view(@project.id, featurestore_id, query, features: feature_schema)
          parsed_json = JSON.parse(json_result)
          expect_status_details(201)

          feature_view_name = parsed_json["name"]
          feature_view_version = parsed_json["version"]
          json_result = get "#{ENV['HOPSWORKS_API']}/project/#{@project.id}/featurestores/#{featurestore_id}/featureview/#{feature_view_name}/version/#{feature_view_version}/preparedstatement?batch=true"
          expect_status_details(200)

          parsed_json = JSON.parse(json_result)
          expect(parsed_json["items"].length).to eql(2)
          expect(parsed_json["items"].first["preparedStatementParameters"].first["index"]).to eql(1)
          expect(parsed_json["items"].first["preparedStatementParameters"].first["name"]).to eql("a_testfeature")
          expect(parsed_json["items"].first["queryOnline"]).to eql("SELECT `fg0`.`a_testfeature1` AS `a_testfeature1`, `fg0`.`a_testfeature` AS `a_testfeature`\nFROM `#{project_name.downcase}`.`#{fg_name}_1` AS `fg0`\nWHERE `fg0`.`a_testfeature` IN ?")
          expect(parsed_json["items"].second["preparedStatementParameters"].first["index"]).to eql(1)
          expect(parsed_json["items"].second["preparedStatementParameters"].first["name"]).to eql("b_testfeature")
          expect(parsed_json["items"].second["queryOnline"]).to eql("SELECT `fg0`.`b_testfeature1` AS `yolo_b_testfeature1`, `fg0`.`b_testfeature` AS `yolo_b_testfeature`\nFROM `#{project_name.downcase}`.`#{fg_name_b}_1` AS `fg0`\nWHERE `fg0`.`b_testfeature` IN ?")
          expect(parsed_json["items"].second["prefix"]).to eql("yolo_")
        end

        it "should fail when calling get serving vector from feature view created from offline fg" do
          # create first feature group
          featurestore_id = get_featurestore_id(@project.id)
          project_name = @project.projectname

          features = [
              {type: "INT", name: "a_testfeature", primary: true},
              {type: "INT", name: "a_testfeature1"},
          ]
          json_result, fg_name_a = create_cached_featuregroup(@project.id, featurestore_id, features: features, featuregroup_name: "test_fg_a_#{short_random_id}", online:true)
          parsed_json = JSON.parse(json_result)
          expect_status_details(201)
          fg_id = parsed_json["id"]
          fg_type = parsed_json["type"]
          # create second feature group
          features = [
              {type: "INT", name: "a_testfeature", primary: true},
              {type: "INT", name: "b_testfeature1"},
          ]
          json_result_b, fg_name_b = create_cached_featuregroup(@project.id, featurestore_id, features: features, featuregroup_name: "test_fg_b_#{short_random_id}", online:false)
          parsed_json_b = JSON.parse(json_result_b)
          expect_status_details(201)
          fg_id_b = parsed_json_b["id"]
          fg_b_type = parsed_json_b["type"]
          # create queryDTO object
          query = {
              leftFeatureGroup: {
                  id: fg_id,
                  type: fg_type
              },
              leftFeatures: [{name: 'a_testfeature1'}],
              joins: [{
                       query: {
                           leftFeatureGroup: {
                               id: fg_id_b,
                               type: fg_b_type
                           },
                           leftFeatures: [{name: 'b_testfeature1'}]
                       }
                  }
              ]
          }

          json_result = create_feature_view(@project.id, featurestore_id, query)
          parsed_json = JSON.parse(json_result)
          expect_status_details(201)

          feature_view_name = parsed_json["name"]
          feature_view_version = parsed_json["version"]
          get "#{ENV['HOPSWORKS_API']}/project/#{@project.id}/featurestores/#{featurestore_id}/featureview/#{feature_view_name}/version/#{feature_view_version}/preparedstatement"
          expect_status_details(400)
        end

        it "should fail when calling get serving vector if a feature group was deleted" do
          # create first feature group
          featurestore_id = get_featurestore_id(@project.id)
          project_name = @project.projectname

          features = [
            {type: "INT", name: "a_testfeature", primary: true},
            {type: "INT", name: "a_testfeature1"},
          ]
          json_result, fg_name_a = create_cached_featuregroup(@project.id, featurestore_id, features: features, featuregroup_name: "test_fg_a_#{short_random_id}", online:true)
          parsed_json = JSON.parse(json_result)
          expect_status_details(201)
          fg_id = parsed_json["id"]
          fg_type = parsed_json["type"]
          # create second feature group
          features = [
            {type: "INT", name: "a_testfeature", primary: true},
            {type: "INT", name: "b_testfeature1"},
          ]
          json_result_b, fg_name_b = create_cached_featuregroup(@project.id, featurestore_id, features: features, featuregroup_name: "test_fg_b_#{short_random_id}", online:false)
          parsed_json_b = JSON.parse(json_result_b)
          expect_status_details(201)
          fg_id_b = parsed_json_b["id"]
          fg_b_type = parsed_json_b["type"]
          # create queryDTO object
          query = {
            leftFeatureGroup: {
              id: fg_id,
              type: fg_type
            },
            leftFeatures: [{name: 'a_testfeature1'}],
            joins: [{
                      query: {
                        leftFeatureGroup: {
                          id: fg_id_b,
                          type: fg_b_type
                        },
                        leftFeatures: [{name: 'b_testfeature1'}]
                      }
                    }
            ]
          }

          json_result = create_feature_view(@project.id, featurestore_id, query)
          parsed_json = JSON.parse(json_result)
          expect_status_details(201)

          feature_view_name = parsed_json["name"]
          feature_view_version = parsed_json["version"]

          # delete the second feature group
          delete_featuregroup_checked(@project.id, featurestore_id, fg_id_b)

          get "#{ENV['HOPSWORKS_API']}/project/#{@project.id}/featurestores/#{featurestore_id}/featureview/#{feature_view_name}/version/#{feature_view_version}/preparedstatement"
          expect_status_details(400)
        end

        it "should fail when calling get serving vector from feature view created from fg without primary key" do
          # create feature group without primary key
          featurestore_id = get_featurestore_id(@project.id)
          project_name = @project.projectname
          features = [
              {type: "INT", name: "a_testfeature"},
              {type: "INT", name: "a_testfeature1"},
          ]
          json_result, fg_name = create_cached_featuregroup(@project.id, featurestore_id, features: features, featuregroup_name: "test_fg_a_#{short_random_id}", online:true)
          parsed_json = JSON.parse(json_result)
          expect_status_details(201)
          fg_id = parsed_json["id"]
          fg_type = parsed_json["type"]

          # create queryDTO object
          query = {
              leftFeatureGroup: {
                  id: fg_id,
                  type: fg_type
              },
              leftFeatures: [{name: 'a_testfeature1'}],
          }

          json_result = create_feature_view(@project.id, featurestore_id, query)
          parsed_json = JSON.parse(json_result)
          expect_status_details(201)

          feature_view_name = parsed_json["name"]
          feature_view_version = parsed_json["version"]
          get "#{ENV['HOPSWORKS_API']}/project/#{@project.id}/featurestores/#{featurestore_id}/featureview/#{feature_view_name}/version/#{feature_view_version}/preparedstatement"
          expect_status_details(400)
        end

        it "should construct prepared statements for snowflake schema" do
          featurestore_id = get_featurestore_id(@project.id)
          featuregroup_suffix = short_random_id
          # should create a snowflake feature view with nested joins
          parsed_json = create_snowflake_fv(@project.id,featurestore_id , featuregroup_suffix)
          feature_view_name = parsed_json["name"]
          feature_view_version = parsed_json["version"]
          # get prepared statements for the feature view
          parsed_query_result = get_prepared_statements(@project.id, featurestore_id, feature_view_name, feature_view_version)
          expect(parsed_query_result["count"]).to eql(3)
          # assert first query
          EXPECTED_QUERY1 = "SELECT `fg0`.`b` AS `b`, `fg0`.`c` AS `c`\n" +
            "FROM `#{@project.projectname}`.`test_fg_1#{featuregroup_suffix}_1`" + " AS `fg0`\n" +
            "WHERE `fg0`.`pk` = ?"
          expect(parsed_query_result["items"].first["queryOnline"]).to eql(EXPECTED_QUERY1)
          # assert the second query
          EXPECTED_QUERY2 = "SELECT `fg1`.`g` AS `g`, `fg1`.`h` AS `h`, `fg1`.`i` AS `i`, `fg2`.`j` AS `j`, `fg2`.`k` AS `k`, `fg2`.`l` AS `l`, `fg0`.`d` AS `d`, `fg0`.`e` AS `e`, `fg0`.`f` AS `f`\n" +
            "FROM `#{@project.projectname}`.`test_fg_2#{featuregroup_suffix}_1` AS `fg0`\n" +
            "LEFT JOIN `#{@project.projectname}`.`test_fg_3#{featuregroup_suffix}_1` AS `fg1` ON `fg0`.`fk_3` = `fg1`.`fk_3`\n" +
            "LEFT JOIN `#{@project.projectname}`.`test_fg_4#{featuregroup_suffix}_1` AS `fg2` ON `fg1`.`fk_4` = `fg2`.`fk_4`\n" +
            "WHERE `fg0`.`fk_2` = ?"
          expect(parsed_query_result["items"].second["queryOnline"]).to eql(EXPECTED_QUERY2)
          # assert the third query
          EXPECTED_QUERY3="SELECT `fg1`.`p` AS `p`, `fg1`.`q` AS `q`, `fg1`.`r` AS `r`, `fg0`.`m` AS `m`, `fg0`.`n` AS `n`, `fg0`.`o` AS `o`\n" +
            "FROM `#{@project.projectname}`.`test_fg_5#{featuregroup_suffix}_1` AS `fg0`\n" +
            "LEFT JOIN `#{@project.projectname}`.`test_fg_6#{featuregroup_suffix}_1` AS `fg1` ON `fg0`.`fk_6` = `fg1`.`fk_6`\n" +
            "WHERE `fg0`.`fk_5` = ?"
          expect(parsed_query_result["items"].third["queryOnline"]).to eql(EXPECTED_QUERY3)
        end

        
        it "should construct prepared statements for snowflake schema-batch mode" do
          featurestore_id = get_featurestore_id(@project.id)
          featuregroup_suffix = short_random_id
          # should create a snowflake feature view with nested joins
          parsed_json = create_snowflake_fv(@project.id,featurestore_id , featuregroup_suffix)
          feature_view_name = parsed_json["name"]
          feature_view_version = parsed_json["version"]
          # get prepared statements for the feature view
          parsed_query_result = get_prepared_statements(@project.id, featurestore_id, feature_view_name, feature_view_version, batch:true)
          expect(parsed_query_result["count"]).to eql(3)
          # assert first query
          EXPECTED_QUERY1 = "SELECT `fg0`.`b` AS `b`, `fg0`.`c` AS `c`, `fg0`.`pk` AS `pk`\n" +
            "FROM `#{@project.projectname}`.`test_fg_1#{featuregroup_suffix}_1`" + " AS `fg0`\n" +
            "WHERE `fg0`.`pk` IN ?"
          expect(parsed_query_result["items"].first["queryOnline"]).to eql(EXPECTED_QUERY1)
          # assert the second query
          EXPECTED_QUERY2 = "SELECT `fg1`.`g` AS `g`, `fg1`.`h` AS `h`, `fg1`.`i` AS `i`, `fg2`.`j` AS `j`, `fg2`.`k` AS `k`, `fg2`.`l` AS `l`, `fg0`.`d` AS `d`, `fg0`.`e` AS `e`, `fg0`.`f` AS `f`, `fg0`.`fk_2` AS `fk_2`\n" +
            "FROM `#{@project.projectname}`.`test_fg_2#{featuregroup_suffix}_1` AS `fg0`\n" +
            "LEFT JOIN `#{@project.projectname}`.`test_fg_3#{featuregroup_suffix}_1` AS `fg1` ON `fg0`.`fk_3` = `fg1`.`fk_3`\n" +
            "LEFT JOIN `#{@project.projectname}`.`test_fg_4#{featuregroup_suffix}_1` AS `fg2` ON `fg1`.`fk_4` = `fg2`.`fk_4`\n" +
            "WHERE `fg0`.`fk_2` IN ?"
          expect(parsed_query_result["items"].second["queryOnline"]).to eql(EXPECTED_QUERY2)
          # assert the third query
          EXPECTED_QUERY3="SELECT `fg1`.`p` AS `p`, `fg1`.`q` AS `q`, `fg1`.`r` AS `r`, `fg0`.`m` AS `m`, `fg0`.`n` AS `n`, `fg0`.`o` AS `o`, `fg0`.`fk_5` AS `fk_5`\n" +
            "FROM `#{@project.projectname}`.`test_fg_5#{featuregroup_suffix}_1` AS `fg0`\n" +
            "LEFT JOIN `#{@project.projectname}`.`test_fg_6#{featuregroup_suffix}_1` AS `fg1` ON `fg0`.`fk_6` = `fg1`.`fk_6`\n" +
            "WHERE `fg0`.`fk_5` IN ?"
          expect(parsed_query_result["items"].third["queryOnline"]).to eql(EXPECTED_QUERY3)
        end

        
        it "should fail when calling get serving vector from feature view created from offline fg-snowflake" do
          # create first feature group
          featurestore_id = get_featurestore_id(@project.id)
          project_name = @project.projectname

          features = [
              {type: "INT", name: "pk_a", primary: true},
              {type: "INT", name: "fk_2"},
          ]
          json_result, fg_name_a = create_cached_featuregroup(@project.id, featurestore_id, features: features, featuregroup_name: "test_fg_a_#{short_random_id}", online:true)
          parsed_json = JSON.parse(json_result)
          expect_status_details(201)
          fg_id = parsed_json["id"]
          fg_type = parsed_json["type"]
          # create second feature group
          features = [
              {type: "INT", name: "fk_2", primary: true},
              {type: "INT", name: "fk_3"},
          ]
          json_result_b, fg_name_b = create_cached_featuregroup(@project.id, featurestore_id, features: features, featuregroup_name: "test_fg_b_#{short_random_id}", online:true)
          parsed_json_b = JSON.parse(json_result_b)
          expect_status_details(201)
          fg_id_b = parsed_json_b["id"]
          fg_b_type = parsed_json_b["type"]
          # fg 3
          features = [
            {type: "INT", name: "fk_3", primary: true},
            {type: "INT", name: "fk_4"},
        ]
        json_result_c, fg_name_c = create_cached_featuregroup(@project.id, featurestore_id, features: features, featuregroup_name: "test_fg_c_#{short_random_id}", online:false)
        parsed_json_c = JSON.parse(json_result_c)
        expect_status_details(201)
        fg_id_c = parsed_json_c["id"]
        fg_c_type = parsed_json_c["type"]
        # create queryDTO object
        query = {
            leftFeatureGroup: {
                id: fg_id,
                type: fg_type
            },
            leftFeatures: [{name: 'fk_2'}],
            joins:[
              {
              query: {
                leftFeatureGroup: {
                  id: fg_id_b,
                  type: fg_b_type
                },
                leftFeatures: [{name: 'fk_3'}],
                joins: [	
                    {
                      query: {
                      leftFeatureGroup: {
                        id: fg_id_c,
                        type: fg_c_type
                      },
                      leftFeatures: [{name: "fk_4"}],
                      joins: [],
                      filter: nil,
                      },
                      on: [],
                      leftOn: [{name: "fk_3"}],
                      rightOn: [{name: "fk_3"}],
                      type: "LEFT"
                  },
                ],
              },          
              on: [],
              leftOn: [{name: "fk_2"}],
              rightOn: [{name: "fk_2"}],
              type: "LEFT"
            }                  
          ],
          filter: nil,
          }

          json_result = create_feature_view(@project.id, featurestore_id, query)
          parsed_json = JSON.parse(json_result)
          expect_status_details(201)

          feature_view_name = parsed_json["name"]
          feature_view_version = parsed_json["version"]
          get "#{ENV['HOPSWORKS_API']}/project/#{@project.id}/featurestores/#{featurestore_id}/featureview/#{feature_view_name}/version/#{feature_view_version}/preparedstatement"
          expect_status_details(400)
        end

        it "should fail when calling get serving vector from feature view created without PK-snowflake" do
          # create first feature group
          featurestore_id = get_featurestore_id(@project.id)
          project_name = @project.projectname

          features = [
              {type: "INT", name: "pk_a", primary: true},
              {type: "INT", name: "fk_2"},
          ]
          json_result, fg_name_a = create_cached_featuregroup(@project.id, featurestore_id, features: features, featuregroup_name: "test_fg_a_#{short_random_id}", online:true)
          parsed_json = JSON.parse(json_result)
          expect_status_details(201)
          fg_id = parsed_json["id"]
          fg_type = parsed_json["type"]
          # create second feature group
          features = [
              {type: "INT", name: "fk_2", primary: true},
              {type: "INT", name: "fk_3"},
          ]
          json_result_b, fg_name_b = create_cached_featuregroup(@project.id, featurestore_id, features: features, featuregroup_name: "test_fg_b_#{short_random_id}", online:true)
          parsed_json_b = JSON.parse(json_result_b)
          expect_status_details(201)
          fg_id_b = parsed_json_b["id"]
          fg_b_type = parsed_json_b["type"]
          # fg 3
          features = [
            {type: "INT", name: "fk_3", primary: false},
            {type: "INT", name: "fk_4"},
        ]
        json_result_c, fg_name_c = create_cached_featuregroup(@project.id, featurestore_id, features: features, featuregroup_name: "test_fg_c_#{short_random_id}", online:true)
        parsed_json_c = JSON.parse(json_result_c)
        expect_status_details(201)
        fg_id_c = parsed_json_c["id"]
        fg_c_type = parsed_json_c["type"]
        # create queryDTO object
        query = {
            leftFeatureGroup: {
                id: fg_id,
                type: fg_type
            },
            leftFeatures: [{name: 'fk_2'}],
            joins:[
              {
              query: {
                leftFeatureGroup: {
                  id: fg_id_b,
                  type: fg_b_type
                },
                leftFeatures: [{name: 'fk_3'}],
                joins: [	
                    {
                      query: {
                      leftFeatureGroup: {
                        id: fg_id_c,
                        type: fg_c_type
                      },
                      leftFeatures: [{name: "fk_4"}],
                      joins: [],
                      filter: nil,
                      },
                      on: [],
                      leftOn: [{name: "fk_3"}],
                      rightOn: [{name: "fk_3"}],
                      type: "LEFT"
                  },
                ],
              },          
              on: [],
              leftOn: [{name: "fk_2"}],
              rightOn: [{name: "fk_2"}],
              type: "LEFT"
            }                  
          ],
          filter: nil,
          }

          create_feature_view(@project.id, featurestore_id, query)
          expect_status_details(400)
        end

      end
    end
  end
end
